// Pawe Pawliski E1ISI //
//                       //
////////////////////////////////////////////////////////////////
//
//            Wzorzec Non-virtual Interface (NVI)
//
////////////////////////////////////////////////////////////////

// Wzorzec NVI pozwala na oddzielenie interfejsu od szczegw implementacji,
// ktre mona modyfikowa dziki dziedziczeniu i polimorfizmowi.
// Gwny pomys opiera si na unikaniu deklarowania funkcji wirtualnych
// jako publicznych.

#include <string>
#include <vector>
#include <exception>

// Przykadowa klasa bazowa CommInterface, implementujca NVI:
class CommInterface {
public:
	// Wszystkie metody publiczne nie s wirtualne, dziki czemu zawsze
	// bd wywoywane metody klasy CommInterface.
	void shutdown () throw (std::exception) {

		// doShutdown to prywatna metoda czysto wirtualna.
		doShutdown ();

		// Dziki temu, e nasz interfejs jest niewirtualny, moemy
		// zagwarantowa pewn funkcjonalno i jednoczenie nie 
		// wymusza na autorze klasy pochodnej jej implementacji, np.
		// moemy zagwarantowa spenienie pewnych warunkw:
		if (not queue.empty()) throw(std::exception());
	}

	// Wzorzec NVI jest silnie zwizany z wzorcem metody szablonu
	// (template method pattern). W skrcie sprowadza si on do
	// umieszczenia w metodzie klasy bazowej schematu algorytmu, ktrego
	// kolejne kroki s wywoaniami prywatnych funkcji wirtualnych.
	// Prostym przykadem moe by ponisze zapytanie:
	std::string& query (std::string &q) {

		validateInput(q);	// krok 1
		send(q);		// krok 2
		std::string& results = recieve();	// krok 3
		validateOutput(results);	// krok 4

		return results;
	}

	// Destruktory s wyjtkiem - musz by zadeklarowane jako wirtualne
	// (konieczne do poprawnego niszczenia obiektw klas pochodnych) oraz
	// posiada dostp publiczny.
	virtual ~CommInterface () {
		// W przypadku klasy CommInterface brak operacji.
	}

protected:
	// Przykadowa kolejka wychodzca.
	std::vector<std::string> queue;
	
private:
	// W tym przykadzie wszystkie funkcje wirtualne z wyjtkiem
	// destruktorw zadeklarowane s jako nie-publiczne.
	// Poniej znajduj si przykadowe "prymitywne" funkcje,
	// reprezentujce szczegy implementacji, o ktre klasa bazowa
	// pozostawia do implementacji klasom pochodnym.

	// Wysanie danych z kolejki i zamknicie poczenia
	virtual void doShutdown () throw () = 0;

	// Wysanie danych z kolejki
	virtual void send (std::string &) = 0;

	// Odebranie danych
	virtual std::string& recieve () = 0;

protected:
	// Zadeklarowanie metod jako protected pozwala klasom pochodnym
	// skorzystanie z gotowych implementacji w klasie bazowej, co moe
	// by bardzo przydatne np. przy serializacji.

	// Walidacja wyjcia od zdalnego serwera
	virtual void validateOutput (std::string s) throw (std::exception) {

		if (s.size() < 1024) throw (std::exception());
	}

	// Walidacja wejcia od uytkownika
	virtual void validateInput (std::string s) throw (std::exception) {

		if (s.size() == 0) throw (std::exception());
	}
};


// Przykadowa konkretne klasa pochodna, implementujca funkcje czysto
// wirtualne.
class CommViaTLS : public CommInterface {
public:
	// W sekcji publicznej nie ma potrzeby implementowania metod klasy
	// bazowej.

private:

	// Przykadowe implementacje.

	virtual void doShutdown () throw () {

		while (!queue.empty()) {
			send(queue.back());
			queue.pop_back();
		}

		// Tu powinno nastpi np. zamknicie sesji TLS.
		// close() ...
	}

	virtual void send (std::string &) {

		// Tu powinno nastpi wysanie danych.
		// write(...)

		queue.pop_back();
	}


	virtual std::string& recieve () {

		// Tu powinno znajdowa si czytanie z deskryptora
		// read(...)
	}
};

// W innych klasach moemy implementowa pozostae funkcje wirtualne - fakt, e
// niektre nie s czysto wirtualne, sygnalizuje domyln implementacj, ktr
// mona zmodyfikowa.
class CommViaDBUS : public CommInterface {
private:
	
	virtual void doShutdown () throw () {
		
		// ...
		// miejsce na konkretn implementacj
	}

	virtual void send (std::string &) {

		// ...
		// miejsce na konkretn implementacj
	}

	virtual std::string& recieve () {

		// ...
		// miejsce na konkretn implementacj
	}
	
	virtual void validateInput (std::string s) throw (std::exception) {

		CommInterface::validateInput(s);
		// ...
		// Tu powinna nastpi dodatkowa walidacja
	}

	virtual void validateOutput (std::string s) throw (std::exception) {

		CommInterface::validateOutput(s);
		// ...
		// Tu powinna nastpi dodatkowa walidacja
	}
};

